* Program Name: utils.prg 
* Purpose: Utility functions & procedures.
* Author: Ralph E. Mead, AKA The Twit, of Georgia-Pacific Corporation
*         (404) 521-4538
* Created on 12/8/1988 at 17:52
*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*
* Procedures/Functions in this file:
*  
* The following procedures are included in this .prg:
*
*   ZPADDFIELD      Add field to creating.dbf
*                   field_name, field_type, field_len, field_dec
*
*   ZPMAKNTX        Check for/make index.  Display message while indexing.
*                   [ N-row, N-col, N-rows-to-clear, ]  C-dbfname , C-ntxname , C-index-expression
*
*   ZPCSV           Write CSV file for spreadsheet import
*                   C-output file name, C-dbf name for header
*
*   ZPTONE          Tone/sound - standard tones & sound speaker output
*                   [ C-option as per CASE statements ] 
*                   'UGET'   'INVALID'   'DONE'   'DONEBIG'
*
*   ZPDC            Display Comment & proceed with no user response
*                   [ N-row, N-col, N-rows to clear, ]  C-comment  
*
*   ZPDCCOL         Display Comment at next column in same row & proceed
*                   C-comment
*
*   ZPDM            Display Message & wait for user to strike key
*                   [ N-row, N-col, N-rows to clear, ]  C-message  
*
*
* The following function are included in this .prg:
*
*   ZFBREAKER()     Return a standard breaker string for use in breaking strings
*
*   ZFESCFLT()      Permit user to escape out of report.
*
*   ZFESCRPT()      Permit user to escape out of filter.
*
*   ZFFILEARR()     Build array of directory info for use as a selection tool
*                   ZFFILEARR('.dbf')        && will produce a PUBLIC array of *.dbf called 'ar_dbfs'
*                   ZFFILEARR('*.*',C-name)  && will produce a PUBLIC array of *.* called C-name
*
*   ZFFILELATE()    Return most-recent of two files (compare directory date/time)
*                   latest_fil = ZFFILELATE(  C-file1    ,  C-file2    )
*                                          ( 'file1.dbf' , 'file2.mem' )
*                                          (  C-memvar1  ,  C-memvar2  )
*
*   ZFMPBOX()       Menu - PopUp Box --  Return first char of option
*                   [ N-toprow, N-leftcol, N-botrow, N-rightcol, ]  C-boxtype, C-title, C-options
*
*   ZFGO()          Get Option from user (single-character validated choice)
*                   [ N-row, N-col, N-rows to clear, ] C-prompt, C-valid choices, C-initial choice
*
*   ZFGV()          Get Variable from user (multi-character non-validated choice)
*                   [ N-row, N-col, N-rows to clear, ]  C-prompt, C-initial value of memvar,  C-picture template
*                     ' ' - Allows any character        '!' - Allows any character and converts to uppercase
*                     '9' - Allows ONLY digits          '#' - Allows only digits and blanks
*                     'A' - Allows ONLY letters
*
*
*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*=-*
*                           && End of .prg header description 


PROCEDURE ZPADDFIELD
* ------------------------------------------------------------------------
* Add field to creating.dbf
* Format: DO ZPADDFIELD WITH  field_name, field_type, field_len, field_dec
*

PRIVATE    field, type, len, dec
PARAMETERS field, type, len, dec

IF PCOUNT() <> 4                 
    dec = 0
ENDIF
IF TYPE('field') <> 'C' .OR. TYPE('type') <> 'C' .OR. TYPE('len') <> 'N' .OR. TYPE('dec') <> 'N'
    RELEASE bad_params
    trash = bad_params  && Force error -- bad parameters
ENDIF
IF .NOT. INT(len) == len .OR. .NOT. INT(dec) == dec
    RELEASE bad_params
    trash = bad_params  && Force error -- bad parameters
ENDIF

APPEND BLANK
REPLACE field_name WITH field, field_type WITH type, field_len WITH len, field_dec WITH IIF(PCOUNT()=4,dec,0)

RETURN                      && End of procedure 


PROCEDURE ZPMAKNTX
* ===========================================================================
* Check for index file.  If not found, create index & display message.
* Format: DO ZPMAKNTX WITH [ N-row,
*                            N-col, 
*                            N-rows-to-clear, ]
*                            C-dbfname , 
*                            C-ntxname , 
*                            C-index-expression
*
* Note that the ZPDC modes of operation apply here:
*   If 3 param, lines 23 & 24 are cleared and message is at 23,00
*   If 6 params, programmer specified:
*       where message is to be displayed,
*       number of lines (starting at message display line) to be cleared,
*       and the message.
*

PRIVATE     param1, param2, param3, param4, param5, param6
PARAMETERS  param1, param2, param3, param4, param5, param6
PRIVATE row, col, rowsclear, dbfname, ntxname, ntxexpr
PRIVATE pcount, index_file, wkarea, nextarea

pcount  = PCOUNT()
DO CASE
    CASE pcount = 6
        dbfname     = UPPER(param4)
        ntxname     = UPPER(param5)
        ntxexpr     = param6
    CASE pcount = 3
        dbfname     = UPPER(param1)
        ntxname     = UPPER(param2)
        ntxexpr     = param3
    OTHERWISE
        RELEASE bad_params
        trash = bad_params  && Force error -- invalid param count
ENDCASE

IF '.' $ ntxname
    index_file = SUBSTR( ntxname, 1, AT('.',ntxname)-1 ) + INDEXEXT()
ELSE
    index_file = ntxname + INDEXEXT()
ENDIF

IF .NOT. FILE( index_file )
    IF pcount = 6
        DO ZPDC WITH param1, param2, param3, 'Indexing ' + dbfname + ' to ' + ntxname + '...'
    ELSE
        DO ZPDC WITH 'Indexing ' + dbfname + ' to ' + ntxname + '...'
    ENDIF

    wkarea = SELECT()           && save current workarea
    SELECT 0                    && select unused workarea
    USE &dbfname.
    INDEX ON &ntxexpr. TO &ntxname.
    USE
    SELECT (wkarea)             && return to current workarea

    IF pcount = 6
        DO ZPDC WITH param1, param2, param3, ''
    ELSE
        DO ZPDC WITH ''
    ENDIF
ENDIF

RETURN                      && End of procedure 


PROCEDURE ZPCSV
* ===========================================================================
* Write CSV file for spreadsheet import.
* Format: DO ZPCSV WITH C-output file name with or without extension,
*                       C-dbf name for header 
*
* Notes:
*   Open dbf and go top before calling this routine.
*   This routine uses ZFESCRPT() to permit user to escape during write.
*   File extension will be .CSV unless output file name has an extension.
*
*

PRIVATE     filespec, headername
PARAMETERS  filespec, headername

PRIVATE fldcount, n, filspec

* --------------------------------------------------------------------------
* establish arrays with required information
fldcount    = FCOUNT()
DECLARE arxz_name[ fldcount ]
DECLARE arxz_type[ fldcount ]
DECLARE arxz_lens[ fldcount ]
DECLARE arxz_decs[ fldcount ]
DECLARE arxz_pict[ fldcount ]
AFIELDS( arxz_name, arxz_type, arxz_lens, arxz_decs )

FOR n = 1 TO fldcount
    DO CASE
        CASE arxz_type[n] = 'C'
                arxz_pict[n] = REPLICATE('x',arxz_lens[n])
        CASE arxz_type[n] = 'N'
            wknum   = INT( arxz_lens[n] + CEILING(arxz_lens[n]/3) - 1 )
            wkchar  = RIGHT('9,999,999,999,999,999,999',wknum)
            IF arxz_decs[n] <> 0
                wkchar = RIGHT(wkchar,LEN(wkchar)-arxz_decs[n]-1)
                wkchar = wkchar + '.' + REPLICATE('9',arxz_decs[n])
                IF wkchar = ','
                    wkchar = SUBSTR(wkchar,2)
                ENDIF
            ENDIF
            arxz_pict[n] = wkchar
        CASE arxz_type[n] = 'D'
            arxz_pict[n] = "MM/DD/YY"
        CASE arxz_type[n] = 'L'
            arxz_pict[n] = "T"
    ENDCASE
NEXT

* --------------------------------------------------------------------------
* begin writing the file
IF '.' $ filespec
    filespec = LTRIM(TRIM( filespec ))
ELSE
    filespec = LTRIM(TRIM( filespec )) + '.CSV'
ENDIF
SET ALTERNATE TO &filespec.
SET ALTERNATE ON
SET CONSOLE OFF

* --------------------------------------------------------------------------
* write description/date/time to first cell on first row
?? ["]
?? headername + ' Detail as of ' + DTOC(DATE()) + ' -- ' + TIME()
?? ["]
?                               && Line feed to end spreadsheet row

* --------------------------------------------------------------------------
* if filter in use, write it on next row
IF '' <> DBFILTER()
    ?? ["]
    ?? 'Filter in use: ' + DBFILTER()
    ?? ["]
    ?                               && Line feed to end spreadsheet row
ENDIF
?                               && Line feed for blank row

* --------------------------------------------------------------------------
* write column widths on next row
FOR n = 1 TO fldcount
    IF n <> 1
        ?? [,]              && delimit fields
    ENDIF
    ?? ["]
    ?? TRIM(LTRIM( STR( IIF( arxz_type[n]='N', LEN(arxz_pict[n])+2, LEN(arxz_pict[n])+1 ) ) ) )
    ?? ["]
NEXT
?                               && Line feed to end spreadsheet row

* --------------------------------------------------------------------------
* write picture statements on next row
FOR n = 1 TO fldcount
    IF n <> 1
        ?? [,]              && delimit fields
    ENDIF
    ?? ["]
    ?? arxz_pict[n]
    ?? ["]
NEXT
?                               && Line feed to end spreadsheet row

* --------------------------------------------------------------------------
* write field names on next row
FOR n = 1 TO fldcount
    IF n <> 1
        ?? [,]              && delimit fields
    ENDIF
    ?? ["]
    ?? TRIM(LTRIM( arxz_name[n] ))
    ?? ["]
NEXT
?                               && Line feed to end spreadsheet row
?                               && Line feed for blank row

* --------------------------------------------------------------------------
* write data to file
DO WHILE .NOT. EOF() .AND. .NOT. ZFESCRPT()
    FOR n = 1 TO fldcount
        IF n <> 1
            ?? [,]              && delimit fields
        ENDIF
        wkfield = arxz_name[n]
        DO CASE
            CASE arxz_type[n] = 'N'
                ?? LTRIM(STR(&wkfield.))
            CASE arxz_type[n] = 'C'
                ?? ["]
                ?? TRIM( &wkfield. )
                ?? ["]
            CASE arxz_type[n] = 'D'
                ?? ["]
                ?? DTOC(&wkfield.)
                ?? ["]
            CASE arxz_type[n] = 'L'
                ?? ["]
                ?? IIF( &wkfield., "T", "F" )
                ?? ["]
        ENDCASE
    NEXT
    ?                               && Line feed to end spreadsheet row
    SKIP
ENDDO

* --------------------------------------------------------------------------
* end writing the file
SET CONSOLE ON
SET ALTERNATE OFF
SET ALTERNATE TO

RETURN                      && End of procedure 


PROCEDURE ZPTONE
* ===========================================================================
*   ZPTONE          Tone/sound - standard tones & sound speaker output
*   
* Format: DO ZPTONE WITH [ C-option as per CASE statements ]
*   
* Notes: TONE() is the standard Clipper speaker function
*        SOUND() is a ProClip library speaker function
*           If ProClip Lib is not being used, change SOUND() to TONE()
*   

PRIVATE param1
PARAMETERS param1

IF PCOUNT() = 0
    SOUND(513,4,.F.)
ELSE
    DO CASE
        CASE UPPER(param1) = 'UGET'         && a utility get is starting
            SOUND( 2096,  1, .F. )
        CASE UPPER(param1) = 'INVALID'      && a validate function detected error
            SOUND( 1048,  3, .F. )
        CASE UPPER(param1) = 'ACK'          && acknowledgement to user
            SOUND(  524,  5, .F. )
        CASE UPPER(param1) = 'DONE'         && a processing routine is done
            SOUND(  262, 50, .F. )
        CASE UPPER(param1) = 'DONEBIG'      && a major processing step is done
            SOUND(  196,140, .F. )
            SOUND(  131,250, .F. )
        OTHERWISE
            RELEASE bad_params
            trash = bad_params  && Force error -- invalid parameter
    ENDCASE
ENDIF

RETURN                      && End of procedure 


PROCEDURE ZPDC
* ===========================================================================
* Display comment & continue processing without any user response.
*
* Format: DO ZPDC WITH [ N-row, 
*                        N-col, 
*                        N-rows to clear, ] 
*                        C-comment  
*
* Two modes of operation:
*   If 1 param, lines 23 & 24 are cleared and comment is at 23,00
*
*   If 4 params, programmer specified:
*       where comment is to be displayed,
*       number of lines (starting at comment display line) to be cleared,
*       and the comment.

PRIVATE    param1, param2, param3, param4
PARAMETERS param1, param2, param3, param4
PRIVATE zrow, zcol, zclear_row, zcomment

* Establish where comment is to go
IF PCOUNT() = 4     && User specified row, col, clear_rows
    zrow        = param1
    zcol        = param2
    zclear_row  = param3
    zcomment    = param4
ELSE                && User specified on comment -- default to row 23
    zrow        = 23
    zcol        = 00
    zclear_row  = 2
    zcomment    = param1
ENDIF

* Clear specified rows
IF zrow+zclear_row-1 > 24
    zclear_row = 25-zrow
ENDIF
@ zrow, 00, zrow+zclear_row-1, 79  BOX '         '

* Display comment
IF TYPE("zcomment") = 'C'
    @ zrow,zcol SAY zcomment
ELSE                    && Programmer specified 2, 3, 5 or more parameters
    RELEASE bad_params
    trash = bad_params  && Force error -- invalid param count
ENDIF

RETURN                      && End of procedure 



PROCEDURE ZPDCCOL
* ===========================================================================
* Display comment at next column/same row without any user response.
*
* Format: DO ZPDCCOL WITH C-comment

PRIVATE    zcomment
PARAMETERS zcomment

@ ROW(),COL() SAY zcomment

RETURN                      && End of procedure 


PROCEDURE ZPDM
* ===========================================================================
* Display message (max of 70 chars), wait for user to hit a key, & clear message.
*
* Format: DO ZPDM WITH [ N-row, 
*                        N-col, 
*                        N-rows to clear, ] 
*                        C-message  
*
* Note that the ZPDC modes of operation apply here:
*   If 1 param, lines 23 & 24 are cleared and message is at 23,00
*   If 4 params, programmer specified:
*       where message is to be displayed,
*       number of lines (starting at message display line) to be cleared,
*       and the message.

PRIVATE    param1, param2, param3, param4
PARAMETERS param1, param2, param3, param4
PRIVATE znothing

* Display message
IF PCOUNT() = 4
    DO ZPDC WITH param1, param2, param3, param4
ELSE
    DO ZPDC WITH param1
ENDIF

* Wait for user to acknowledge
SET CONSOLE OFF
WAIT to znothing
SET CONSOLE ON

* Clear message
IF PCOUNT() = 4
    DO ZPDC WITH param1, param2, param3, ''
ELSE
    DO ZPDC WITH '' 
ENDIF

RETURN                      && End of procedure 


FUNCTION zfbreaker
* ===========================================================================
* Return a standard breaker string for use in breaking strings
* As a rule, use this as the default value for the 'm_abreak' public memvar
* because it is the default value assumed by several generic procs/funcs
* which contain a code section like this:
*   IF TYPE('m_abreak') = 'U'
*       breaker = CHR(219)
*   ELSE
*       breaker = ALLTRIM(ZFBREAKER())  && ZFBREAKER() is the array field spacer (field break key)
*   ENDIF
PRIVATE zbreaker
zbreaker = '  '
RETURN zbreaker                     && End of function 


FUNCTION zfescflt
* ===========================================================================
* Permit user to escape out of a filter search.
*
* Format: SET FILTER TO ZFESCFLT() + &zf_str.
*
* Function also sets the value of a 'zfescflt' memvar to .T. for possible use in reports.
*

IF INKEY() = 27
    DO ZPTONE WITH 'ack'    && sound acknowledgement tone
    IF TYPE('zzrec_no') = 'N'       && if memvar for current record number, go to it
        GO zzrec_no
    ENDIF
    zfescrpt = .T.
    RETURN (.T.)                    && NOTE RETURN ****
ENDIF

RETURN  (.F.)                       && End of function 


FUNCTION zfescrpt
* ===========================================================================
* Permit user to escape out of a report.
* If ESC key is hit, go to bottom of file and return a .T..  Else, return a .F..
*
* Format: DO WHILE .NOT. EOF() .AND. ZFESCRPT()
*
* Function also sets the value of a 'zfescrpt' memvar to .T. for possible use in reports.
*

IF INKEY() = 27
    DO ZPTONE WITH 'ack'    && sound acknowledgement tone
    GO BOTTOM
    zfescrpt = .T.
    RETURN (.T.)                    && NOTE RETURN ****
ENDIF

RETURN  (.F.)                       && End of function 


FUNCTION zffilearr
* ===========================================================================
* Build array of file directory information for use as a selection tool
* 
* Format: ZFFILEARR('.dbf')        && will produce a PUBLIC array of *.dbf called 'ar_dbfs'
*         ZFFILEARR('*.*',C-name)  && will produce a PUBLIC array of *.* called C-name
* 
* Return: L- .T. means array was successfully built and has at least one element
*            .F. means array was not built because no such files in current directory
* 
* 

PRIVATE     param1, param2
PARAMETERS  param1, param2
PRIVATE     ext, arrname, fil_count, wk_file, wk_size, wk_date, wk_time, wk_attr, n

IF PCOUNT() = 1
    ext         = ALLTRIM(param1)
    arrname     = 'ar_' +SUBSTR(ext,2) +'s'
ELSE
    ext         = ALLTRIM(param1)
    arrname     = ALLTRIM(param2)
ENDIF    

IF ext = '*.*'
    fil_count   = ADIR(ext) +1      && Add one to permit ADEL() to work properly
ELSE
    fil_count   = ADIR('*'+ext) +1  && Add one to permit ADEL() to work properly
ENDIF
IF fil_count == 1
    RETURN .F.                  && NOTE RETURN
ENDIF

IF TYPE('m_abreak') = 'U'
    PRIVATE  m_abreak
    m_abreak = ZFBREAKER()      && Array field spacer (field break key)
ENDIF

PUBLIC  &arrname.[ fil_count ]
DECLARE wk_file[ fil_count ]
DECLARE wk_size[ fil_count ]
DECLARE wk_date[ fil_count ]
DECLARE wk_time[ fil_count ]
DECLARE wk_attr[ fil_count ]
IF ext = '*.*'
    ADIR( ext, wk_file, wk_size, wk_date, wk_time, wk_attr )
ELSE
    ADIR( '*'+ext, wk_file, wk_size, wk_date, wk_time, wk_attr )
ENDIF
FOR n = 1 TO fil_count-1    && use count-1 because last elements are null -- see fil_count above
    &arrname.[n] = wk_file[n] +SPACE(12-LEN(wk_file[n])) +m_abreak+ STR(wk_size[n],7) +m_abreak+ DTOC(wk_date[n]) +m_abreak+ wk_time[n] +m_abreak+ wk_attr[n]
NEXT

ASORT( &arrname., 1, fil_count-1 )

RELEASE     param1, param2, ext, arrname, fil_count, wk_size, wk_date, wk_time, wk_attr, n

RETURN (.T.)                        && End of function 


FUNCTION ZFFILELATE
* ===========================================================================
* Determine most recent of two files.
* Format: latest_fil = ZFFILELATE(  C-file1    ,  C-file2    )
*                                ( 'file1.dbf' , 'file2.mem' )
*                                (  C-memvar1  ,  C-memvar2  )
*
* Returns: C-'NO FILES'  if neither file1 nor file2 is found
*          C-'NO LATEST' if both files are have same date/time
*          C-file1       if it is later than file2 or if file2 does not exist
*          C-file2       if it is later than file1 or if file1 does not exist
*

PRIVATE    file1, file2, isfile1, isfile2, trash, file1dt, file2dt
PARAMETERS file1, file2

* Verify both files exist
isfile1     = FILE(file1)
isfile2     = FILE(file2)
DO CASE
    CASE isfile1 .AND. isfile2
        * No action - continue processing
    CASE isfile1 .AND. .NOT. isfile2
        RETURN ( file1 )                            && NOTE RETURN *******
    CASE .NOT. isfile1 .AND. isfile2
        RETURN ( file2 )                            && NOTE RETURN *******
    OTHERWISE
        RETURN ( 'NO FILES' )                           && NOTE RETURN *******
ENDCASE

* Commented out section is for use if ProClip Lib is not available.
* Build directory array & get date/time strings accurate to HH:M = 10-minute intervals
*PRIVATE    file1, file2, isfile1, isfile2, no_files, file1el, file2el, file1dt, file2dt
*no_files    = ADIR("*.*")   && get element count
*dummy       = ""            && don't want file sizes
*DECLARE fnames[no_files], fdates[no_files], ftimes[no_files]
*ADIR( "*.*", fnames, dummy, fdates, ftimes )
*
*file1el     = ASCAN( fnames, file1 )
*file2el     = ASCAN( fnames, file2 )
*
*file1dt     = CDTOS( DTOC( fdates[file1el] ) , "MM/DD/YY" ) + SUBSTR( ftimes[file1el] ,1 ,5)
*file2dt     = CDTOS( DTOC( fdates[file2el] ) , "MM/DD/YY" ) + SUBSTR( ftimes[file2el] ,1 ,5)
*
*RELEASE fnames, fdates, ftimes

* The next 4 lines of code assume ProClip Lib is available
* Build two date/time strings with a YYYYMMDDHH:MM:SS format
trash   = SDIR( file1, 'first')         && SDIR is a ProClip function
file1dt = DTOS( SDIR( 'date' ) ) + SDIR( 'time' )
trash   = SDIR( file2, 'first')         && SDIR is a ProClip function
file2dt = DTOS( SDIR( 'date' ) ) + SDIR( 'time' )

DO CASE
    CASE file1dt == file2dt
        RETURN ("NO LATEST")                    && NOTE RETURN *******
    CASE file1dt > file2dt
        RETURN (file1)                          && NOTE RETURN *******
    CASE file2dt > file1dt
        RETURN (file2)                          && NOTE RETURN *******
ENDCASE

RETURN ("NO LATEST")                        && End of function 


FUNCTION ZFMPBOX
* ===========================================================================
* Menu - Pop-up Box Menu 
*
* Format: ZFMPBOX( [ N-toprow, 
*                    N-leftcol, 
*                    N-botrow, 
*                    N-rightcol, ]
*                    C-boxtype, 
*                    C-title, 
*                    C-options
*                  [ L-screen save ] )
* 
*       toprow      OPTIONAL - top row of box
*       leftcol     OPTIONAL - left column of box
*       botrow      OPTIONAL - bottom row of box
*       rightcol    OPTIONAL - right column of box
*               If coodinates are not specified, box will be centered on
*                   screen using rows 3 thru 22.
*               Partial use of defaults can be achieved by assigning
*               values of 99 to certain coordinates:
*                   xx,99,xx,99 -- box will be centered horizontally.
*                   99,xx,99,xx -- box will be centered vertically.
*                   xx,xx,99,99 -- box will be positioned from 
*                       upper-left coordinates if possible.
* 
*       boxtype     Two characters:
*                       'DD'  top/bottom DOUBLE-line, sides DOUBLE-line
*                       'SS'  top/bottom single-line, sides single-line
*                       'DS'  top/bottom DOUBLE-line, sides single-line
*                       'SD'  top/bottom single-line, sides DOUBLE-line
* 
*       title       Title to be centered on top line of box.
*                       Specify 'none' if no title desired.  Case insensitive.
* 
*       options     String with menu items separated by m_abreak memvar or 
*                   by the ZFBREAKER() value.
*                       Each item must be less than 70 chars.
*                       First character of each string must be unique
*
*       screensave  Logical - .T. (default) save screen before/restore screen after
*                             .F. don't save screen before/restore screen after
*
*
* Returns:  First letter in menu item selected.
*           If parameters incorrectly specified & cannot handle,
*               returns an undeclared memvar to generate run error.
*
* Comments: Routine will save screen before/restore screen after
*               (Screen memvar is z_m_pbox)
*
* Example:
*   ztitle = 'SELECT BROWSE DESIRED'
*   zoptions = ''
*   zoptions = zoptions + "1. Balances - Beginning & Ending         " +m_abreak
*   zoptions = zoptions + "2. Additions                             " +m_abreak
*   zoptions = zoptions + "Quit ----------------------------------- " +m_abreak
*   zwant_to = ZFMPBOX( 'DD', ztitle , zoptions, .F. )
*   IF zwant_to = 'Q'
*       EXIT                                     && Note EXIT
*   ENDIF
*   DO CASE
*       CASE zwant_to = '1' ...
*       CASE zwant_to = '2' ...
*   ENDCASE
*

PRIVATE    param1, param2,  param3, param4,   param5,  param6, param7, param8
PARAMETERS param1, param2,  param3, param4,   param5,  param6, param7, param8

PRIVATE    toprow, leftcol, botrow, rightcol, boxtype, title,  options

PRIVATE    coord_spec, opt_long, opt_count, opt_poss, opt_which
PRIVATE    left_box, right_box, top_box, bot_box, save_it
PRIVATE    z_m_pbox, boxstring, row, count, breaker, menu_to, choice, finalok
FOR count = 1 TO 25     && Make private memvars 'mess1' thru 'mess25'
    opt_which = ALLTRIM(STR(count))     && assume 25 is most choices will ever face
    PRIVATE mess&opt_which.
NEXT
        
IF TYPE('param1') = 'N'
    coord_spec  = .T.
    toprow      = param1
    leftcol     = param2
    botrow      = param3
    rightcol    = param4
    boxtype     = UPPER( param5 )
    title       = UPPER( param6 )
    options     = param7
    IF TYPE('param8') = 'U'
        save_it = .F.
    ELSE
        save_it = param8
    ENDIF
ELSE
    coord_spec  = .F.
    toprow      = 99
    leftcol     = 99
    botrow      = 99
    rightcol    = 99
    boxtype     = UPPER( param1 )
    title       = UPPER( param2 )
    options     = param3
    IF TYPE('param4') = 'U'
        save_it = .F.
    ELSE
        save_it = param4
    ENDIF
ENDIF

* Isolate the options. Establish number of options & longest options during this.
opt_long    = 0         && Longest option
opt_count   = 0         && Number of options
opt_poss    = ''        && First char of each option
options     = ALLTRIM( options )
IF TYPE('m_abreak') = 'U'
    breaker = ALLTRIM(ZFBREAKER())  && ZFBREAKER() is the array field spacer (field break key)
ELSE
    breaker = ALLTRIM(m_abreak)
ENDIF
DO WHILE LEN( options )  > LEN(breaker)
    opt_count   = opt_count + 1
    opt_poss    = opt_poss + SUBSTR(options,1,1)
    opt_which   = ALLTRIM(STR(opt_count))
    mess&opt_which. = ' '+ ALLTRIM(SUBSTR( options , 1 , AT(breaker,options)-1 )) + ' '
    options         = ALLTRIM( SUBSTR( options , AT(breaker,options)+1 ) )
    IF LEN( mess&opt_which. ) > opt_long
        opt_long =  LEN( mess&opt_which. )
    ENDIF
ENDDO

* Verify longest option is longer than title.
IF title = 'NONE'
    title = ''
ENDIF
IF LEN( title ) + 2 > opt_long
    opt_long = LEN( title )
ENDIF

* Return if longest option is greater than 72 chars.
IF opt_long > 72  
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- A message is too long
ENDIF

* Establish spacing on screen assuming default positioning
left_box    = INT( ( 79 - opt_long + 2 ) / 2 )
right_box   = left_box + opt_long + 2
top_box     = 3 + INT( (22-opt_count)/2 ) - 2
bot_box     = top_box + opt_count + 1

* Adjust default spacing for variety of possible positioning parameters.
IF coord_spec
    DO CASE
        * Center vertically, if possible
        CASE toprow = 99 .AND. botrow = 99
            IF leftcol <> 99 .AND. rightcol <> 99 .AND. rightcol-leftcol - 2 >= opt_long .AND. rightcol <= 79
                left_box    = leftcol
                right_box   = rightcol
            ELSE
                RELEASE bad_params
                RETURN bad_params   && NOTE RETURN -- Something wrong
            ENDIF

        * Center horizontally, if possible
        CASE leftcol = 99 .AND. rightcol = 99
            IF toprow <> 99 .AND. botrow <> 99 .AND. botrow-toprow - 1 >= opt_count .AND. botrow <= 25
                top_box     = toprow
                bot_box     = botrow
            ELSE
                RELEASE bad_params
                RETURN bad_params   && NOTE RETURN -- Something wrong
            ENDIF

        * Position from upper-left coordinates specified, if possible
        CASE toprow <> 99 .AND. leftcol <> 99
            IF botrow = 99 .AND. rightcol = 99 .AND. toprow + opt_count + 1 <= 24 .AND. leftcol + opt_long + 2 <= 79
                top_box     = toprow
                bot_box     = toprow + opt_count + 1
                left_box    = leftcol
                right_box   = leftcol + opt_long + 2
            ELSE
                RELEASE bad_params
                RETURN bad_params   && NOTE RETURN -- Something wrong
            ENDIF
    ENDCASE
ENDIF

* Draw the box
DO CASE
    CASE boxtype = 'DD'
        boxstring = CHR(201) + CHR(205) + CHR(187) + CHR(186) + CHR(188) + CHR(205) + CHR(200) + CHR(186) + ' '
    CASE boxtype = 'SS'
        boxstring = CHR(218) + CHR(196) + CHR(191) + CHR(179) + CHR(217) + CHR(196) + CHR(192) + CHR(179) + ' '
    CASE boxtype = 'DS'
        boxstring = CHR(213) + CHR(205) + CHR(184) + CHR(179) + CHR(190) + CHR(205) + CHR(212) + CHR(179) + ' '
    CASE boxtype = 'SD'
        boxstring = CHR(214) + CHR(196) + CHR(183) + CHR(186) + CHR(189) + CHR(196) + CHR(211) + CHR(186) + ' '
    OTHERWISE
        RELEASE bad_params
        RETURN bad_params   && NOTE RETURN -- Invalid box type
ENDCASE

IF save_it
    SAVE SCREEN TO z_m_pbox
ENDIF
@ top_box,left_box,bot_box,right_box BOX boxstring

* Write the title
IF '' <> title
    @ top_box, left_box + ( opt_long - LEN(title) ) /2 +1 SAY ' '+ title +' '
ENDIF

* Present the title and get the option
choice  = ' '
menu_to = 1
finalok = .F.
DO WHILE .NOT. finalok
    row     = top_box
    SET CURSOR OFF
    FOR count = 1 TO opt_count
        opt_which   = ALLTRIM(STR(count))
        @ row+1,left_box+2  PROMPT  mess&opt_which. + SPACE(opt_long-LEN(mess&opt_which.)-1)
        row     = row + 1
    NEXT
    MENU TO menu_to
    SET CURSOR ON

    * If choice is not ESC key (menu_to = 0), pick option based on 
    * position of menu_to in possible options string.
    IF menu_to <> 0 .AND. menu_to <= LEN(opt_poss)
        choice = SUBSTR(opt_poss,menu_to,1)
        finalok = .T.
    ELSE                && Choice is ESC key, reset choice & allow to loop.
        menu_to = 1
        finalok = .F.
    ENDIF
ENDDO

IF save_it
    RESTORE SCREEN FROM z_m_pbox
ENDIF

RETURN (choice)                 && End of function 


FUNCTION ZFGO
* ===========================================================================
* Get option from user: displaying prompt and validate user's choice
*
* Format: memvar = ZFGO( [ N-row, 
*                          N-col, 
*                          N-rows to clear, ] 
*                          C-prompt, 
*                          C-valid choices, 
*                          C-initial choice )
*
* Return: C-choice selected.  1-byte uppercase.
*
* Note that the ZPDC modes of operation apply here:
*   If 3 params, lines 23 & 24 are cleared and prompt is at 23,00
*   If 6 params, programmer specified:
*       where prompt is to be displayed,
*       number of lines (starting at prompt display line) to be cleared.
*
* Note that no READ is used in this function.

PRIVATE     param1, param2, param3, param4, param5, param6
PARAMETERS  param1, param2, param3, param4, param5, param6
PRIVATE disp_row, disp_col, prompt, valchoices, choice, initchoice
PRIVATE reg_color, flip_color, inchoice, error

IF PCOUNT() <> 3 .AND. PCOUNT() <> 6
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- Invalid param count
ENDIF

* Clear screen and display prompt
IF PCOUNT() = 6
    disp_row    = param1
    disp_col    = param2
    prompt      = param4
    valchoices  = param5
    initchoice  = param6
    DO ZPDC WITH param1, param2, param3, prompt     && Display prompt
ELSE
    disp_row    = 23
    disp_col    = 00
    prompt      = param1
    valchoices  = param2
    initchoice  = param3
    DO ZPDC WITH prompt                             && Display prompt
ENDIF    

* Display current choice in highlight colors
reg_color   = SETCOLOR()                            && current colors
flip_color  = SUBSTR( reg_color, AT(',',reg_color)+1 )      && get color-string starting with enhanced
flip_color  = SUBSTR( flip_color, 1, AT(',',flip_color)-1 ) && trim other params off end of color-string
SET COLOR TO &flip_color.
@ disp_row,disp_col + LEN(prompt) + 2 SAY UPPER(initchoice)

* Get user's selection, validating against valchoices
DO ZPTONE WITH 'uget'   && sound get tone
error  = .F.
choice      = '~'
DO WHILE .NOT. choice $ valchoices

    IF error = .T.
        DO ZPTONE WITH 'invalid'  && sound error tone
    ELSE
        error = .T.
    ENDIF

    @ disp_row,disp_col + LEN(prompt) + 2 SAY ''
    inchoice  = INKEY(0)

    IF inchoice   = 13      && user hit enter key
        choice  = UPPER(initchoice)
    ELSE
        choice  = UPPER( CHR(inchoice) )
    ENDIF
ENDDO

@ disp_row,disp_col + LEN(prompt) + 2 SAY choice
SET COLOR TO &reg_color.

RETURN  choice              && End of function 


FUNCTION ZFGV
* ===========================================================================
* Get variable from user: displaying prompt
*
* Format: memvar = ZFGV( [ N-row, 
*                          N-col, 
*                          N-rows to clear, ] 
*                          C-prompt, 
*                          C-initial value of memvar,
*                          C-picture template:
*                              ' ' - Allows any character
*                              '9' - Allows ONLY digits
*                              'A' - Allows ONLY letters
*                              '!' - Allows any character and converts to uppercase
*                              '#' - Allows only digits and blanks
*                              Examples:
*                                  '!AA ##' is full spec for template -- MUST be same length as initval
*                                  'AAA999' is full spec for template -- MUST be same length as initval
*                                  ' ' will convert to REPLICATE(' ',LEN(initval))
*                                  '9' will convert to REPLICATE('9',LEN(initval))
*                                  'A' will convert to REPLICATE('A',LEN(initval))
*                                  '!' will convert to REPLICATE('!',LEN(initval))
*                                  '#' will convert to REPLICATE('#',LEN(initval))
*
* Return: C-string entered.  Length same as initial value.
*
* Validation is limited to ensuring the returned value:
*   --  has no ASCII control chars ( 1 thru 31 )
*   --  has no ASCII extended chars ( 127 - 255 )
*   --  is not blank
*
* Note that the ZPDC modes of operation apply here:
*   If 3 params, lines 23 & 24 are cleared and prompt is at 23,00
*   If 6 params, programmer also specified:
*       where prompt is to be displayed,
*       number of lines (starting at prompt display line) to be cleared.
*
* Note that READ is used in this function.
*

PRIVATE    param1, param2, param3, param4, param5, param6
PARAMETERS param1, param2, param3, param4, param5, param6
PRIVATE disp_row, disp_col, prompt, choice, initval, template
PRIVATE testplate, getval, testval, n, error

IF PCOUNT() <> 3 .AND. PCOUNT() <> 6
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- Invalid param count
ENDIF

* Clear screen and display prompt
IF PCOUNT() = 6
    disp_row    = param1
    disp_col    = param2
    prompt      = param4
    initval     = param5
    template    = UPPER(param6)
    DO ZPDC WITH param1, param2, param3, prompt         && Display prompt
ELSE
    disp_row    = 23
    disp_col    = 00
    prompt      = param1
    initval     = param2
    template    = UPPER(param3)
    DO ZPDC WITH prompt                                 && Display prompt
ENDIF    

* Establish & validate template
IF LEN(template) = 0                && ensure template was specified
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- Invalid template
ENDIF
IF LEN(template) = 1
    template = REPLICATE(template,LEN(initval))   && Set up picture statement
ENDIF
IF LEN(initval) <> LEN(template)    && ensure template and initval are same length
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- Invalid template
ENDIF
testplate = template
testplate = CHRSWAP( testplate, '9' )   && delete all 9's out of testplate
testplate = CHRSWAP( testplate, 'A' )   && delete all A's out of testplate
testplate = CHRSWAP( testplate, '!' )   && delete all !'s out of testplate
testplate = CHRSWAP( testplate, '#' )   && delete all #'s out of testplate
IF LEN(ALLTRIM( testplate )) > 0
    RELEASE bad_params
    RETURN bad_params   && NOTE RETURN -- Invalid template
ENDIF

* Get user's input
DO ZPTONE WITH 'uget'   && sound get tone
error  = .F.
DO WHILE .T.

    IF error = .T.
        DO ZPTONE WITH 'invalid'  && sound error tone
    ELSE
        error = .T.
    ENDIF

    getval = initval
    SET CONFIRM ON
    @ disp_row, disp_col + LEN(prompt) + 2 GET getval PICTURE template
    SET CONFIRM OFF
    READ

    testval = getval
    FOR n =   1 TO 31       && trim out chars   1 thru  31 -- ASCII ctrl chars
        testval = CHRSWAP( testval, CHR(n) )
    NEXT
    FOR n = 127 TO 255      && trim out chars 127 thru 255 -- ASCII extend chars
        testval = CHRSWAP( testval, CHR(n) )
    NEXT

    IF testval == getval .AND. '' <> TRIM(getval)   && no chars trimmed out and not blank
        EXIT                                        && passed test, so exit loop
    ENDIF

ENDDO

RETURN getval               && End of function 
